from . import util
import numpy as np

# Fast iterative method to find the MAP scores of the logistic_prior model


# TODO: Implement as a wrapper of a c++ function (and also make sure the rest of pipeline is efficient)
def logistic_prior_MAP(match_list, max_iters=1000, threshold=1e-8):
    """Get MAP estimates of the logistic_prior model with rapid iterative methods.

    :param match_list: List of matches, each represented by a dict of the winner and loser.
    :type match_list: list
    :param max_iters: Maximum number of iterations to perform, defaults to 1000
    :type max_iters: int, optional
    :param threshold: Average change in scores over an iteration at which to stop iterations, defaults to 1e-8
    :type threshold: float, optional
    :return: pandas DataFrame of labels, scores, and errors (inferred by change over an iteration).
    :rtype: DataFrame
    """
    # Convert match_list to a sparse format of neighbors for iteration
    string_indices_dict = util.get_string_indices_dict(match_list)
    n = len(string_indices_dict)

    # TODO: at least take advantage of the sparsity of A
    A = np.zeros((n, n))
    for match in match_list:
        winner_index = string_indices_dict[match["winner"]]
        loser_index = string_indices_dict[match["loser"]]
        A[winner_index, loser_index] += 1

    # Initialize choices of strengths (pi[i] = exp(s[i]))
    pis = np.ones(n)

    # Will estimate the errors in s simply by the change in each iteration
    estimated_s_errors = np.ones(n)

    # Implement the iterative method of Eq. (27) in Newman 2022
    # to find MAP scores for the Bradley-Terry with logistic prior
    # This is approximately our model with alpha = 0, beta = 2.56
    for iter in range(max_iters):
        # Important to perform the updates asynchronously,
        # the synchronous version does not necessarily converge
        for i in range(n):
            # 1/(pi_i + 1) + sum_j A_{ij} pi_j/(pi_i + pi_j)
            numerator = 1 / (pis[i] + 1) + np.dot(A[i, :], pis / (pis[i] + pis))
            # 1/(pi_i + 1) + sum_j A_{ji}/(pi_i + pi_j)
            denominator = 1 / (pis[i] + 1) + np.dot(A[:, i], 1 / (pis[i] + pis))
            new_pi = numerator / denominator

            # Estimate the error in s
            s_change = (new_pi - pis[i]) / pis[
                i
            ]  # Ds = log(pi + Dpi) - log(pi) \approx Dpi/pi
            pis[i] = new_pi
            estimated_s_errors[i] = np.abs(s_change)

        # Check if the average estimated error is below threshold
        RMS_s_error = np.sqrt(np.mean(estimated_s_errors**2))
        if RMS_s_error < threshold:
            break

    # Convert back to scores
    estimated_s = np.log(pis)

    return estimated_s, estimated_s_errors



from ranking.results import *
from ranking.parsing import read_match_list_from_adj_matrix

# Test against results obtained with Mathematica via
# numerical optimization/integration on a small test case
# NOTE: these tests may randomly fail since they assess a MCMC process

# SCORE ESTIMATES


def test_logisticPriorMAP():
    # How many times away from the claimed error we will allow
    deviations_threshold = 5

    # Load test information
    match_list = read_match_list_from_adj_matrix(
        "tests/test_case_files/adj_matrix_1.txt"
    )
    f_in = open("tests/test_case_files/logisticPriorMAP_1.txt", "r")
    true_scores = []
    for line in f_in.readlines():
        true_scores.append(float(line))
    n = len(true_scores)

    # Call function
    scores_df = scores(match_list, force_mode="MAP")

    # Compare results
    for i in range(n):
        player_label = f"player_{i}"
        est_score = scores_df.loc[scores_df["label"] == player_label, "score"].values[0]
        est_score_error = scores_df.loc[
            scores_df["label"] == player_label, "score_error"
        ].values[0]
        # Checking that the estimated error is realistic for each score
        if np.abs(true_scores[i] - est_score) > deviations_threshold * est_score_error:
            print(f"Failed on {i}")
            print("True scores:")
            print(true_scores)
            print("Results:")
            print(scores_df)
            assert False

    # All good
    assert True


def test_logisticPriorEs():
    # How many times away from the claimed error we will allow
    deviations_threshold = 5

    # Load test information
    match_list = read_match_list_from_adj_matrix(
        "tests/test_case_files/adj_matrix_1.txt"
    )
    f_in = open("tests/test_case_files/logisticPriorEs_1.txt", "r")
    true_scores = []
    for line in f_in.readlines():
        true_scores.append(float(line))
    n = len(true_scores)

    # Call function
    scores_df = scores(match_list, force_mode="average", model_name="logistic_prior")

    # Compare results
    for i in range(n):
        player_label = f"player_{i}"
        est_score = scores_df.loc[scores_df["label"] == player_label, "score"].values[0]
        est_score_error = scores_df.loc[
            scores_df["label"] == player_label, "score_error"
        ].values[0]
        # Checking that the estimated error is realistic for each score
        if np.abs(true_scores[i] - est_score) > deviations_threshold * est_score_error:
            print(f"Failed on {i}")
            print("True scores:")
            print(true_scores)
            print("Results:")
            print(scores_df)
            assert False

    # All good
    assert True


def test_depthOnlyEs():
    # How many times away from the claimed error we will allow
    deviations_threshold = 5

    # Load test information
    match_list = read_match_list_from_adj_matrix(
        "tests/test_case_files/adj_matrix_1.txt"
    )
    f_in = open("tests/test_case_files/depthOnlyEs_1.txt", "r")
    true_scores = []
    for line in f_in.readlines():
        true_scores.append(float(line))
    n = len(true_scores)

    # Call function
    scores_df = scores(match_list, force_mode="average", model_name="depth_only")

    # Compare results
    for i in range(n):
        player_label = f"player_{i}"
        est_score = scores_df.loc[scores_df["label"] == player_label, "score"].values[0]
        est_score_error = scores_df.loc[
            scores_df["label"] == player_label, "score_error"
        ].values[0]
        # Checking that the estimated error is realistic for each score
        if np.abs(true_scores[i] - est_score) > deviations_threshold * est_score_error:
            print(f"Failed on {i}")
            print("True scores:")
            print(true_scores)
            print("Results:")
            print(scores_df)
            assert False

    # All good
    assert True


def test_luckOnlyEs():
    # How many times away from the claimed error we will allow
    deviations_threshold = 5

    # Load test information
    match_list = read_match_list_from_adj_matrix(
        "tests/test_case_files/adj_matrix_1.txt"
    )
    f_in = open("tests/test_case_files/luckOnlyEs_1.txt", "r")
    true_scores = []
    for line in f_in.readlines():
        true_scores.append(float(line))
    n = len(true_scores)

    # Call function
    scores_df = scores(match_list, force_mode="average", model_name="luck_only")

    # Compare results
    for i in range(n):
        player_label = f"player_{i}"
        est_score = scores_df.loc[scores_df["label"] == player_label, "score"].values[0]
        est_score_error = scores_df.loc[
            scores_df["label"] == player_label, "score_error"
        ].values[0]
        # Checking that the estimated error is realistic for each score
        if np.abs(true_scores[i] - est_score) > deviations_threshold * est_score_error:
            print(f"Failed on {i}")
            print("True scores:")
            print(true_scores)
            print("Results:")
            print(scores_df)
            assert False

    # All good
    assert True


def test_depthAndLuckEs():
    # How many times away from the claimed error we will allow
    deviations_threshold = 5

    # Load test information
    match_list = read_match_list_from_adj_matrix(
        "tests/test_case_files/adj_matrix_1.txt"
    )
    f_in = open("tests/test_case_files/depthAndLuckEs_1.txt", "r")
    true_scores = []
    for line in f_in.readlines():
        true_scores.append(float(line))
    n = len(true_scores)

    # Call function
    scores_df = scores(match_list, force_mode="average", model_name="depth_and_luck")

    # Compare results
    for i in range(n):
        player_label = f"player_{i}"
        est_score = scores_df.loc[scores_df["label"] == player_label, "score"].values[0]
        est_score_error = scores_df.loc[
            scores_df["label"] == player_label, "score_error"
        ].values[0]
        # Checking that the estimated error is realistic for each score
        if np.abs(true_scores[i] - est_score) > deviations_threshold * est_score_error:
            print(f"Failed on {i}")
            print("True scores:")
            print(true_scores)
            print("Results:")
            print(scores_df)
            assert False

    # All good
    assert True


# PROBABILITY ESTIMATES


def test_logisticPriorMAP_probability():
    # How many times away from the claimed error we will allow
    deviations_threshold = 5

    # Load test information
    match_list = read_match_list_from_adj_matrix(
        "tests/test_case_files/adj_matrix_1.txt"
    )
    f_in = open("tests/test_case_files/logisticPriorP_1.txt", "r")
    line = f_in.readline()
    true_prob = float(line)

    # Call function
    prob, prob_error = probability(match_list, "player_0", "player_1", force_mode="MAP")

    # Compare results
    if np.abs(true_prob - prob) > deviations_threshold * prob_error:
        assert False

    # All good
    assert True


def test_logisticPrior_probability():
    # How many times away from the claimed error we will allow
    deviations_threshold = 5

    # Load test information
    match_list = read_match_list_from_adj_matrix(
        "tests/test_case_files/adj_matrix_1.txt"
    )
    f_in = open("tests/test_case_files/logisticPriorEP_1.txt", "r")
    line = f_in.readline()
    true_prob = float(line)

    # Call function
    prob, prob_error = probability(
        match_list,
        "player_0",
        "player_1",
        force_mode="average",
        model_name="logistic_prior",
    )

    # Compare results
    if np.abs(true_prob - prob) > deviations_threshold * prob_error:
        assert False

    # All good
    assert True


def test_depthOnly_probability():
    # How many times away from the claimed error we will allow
    deviations_threshold = 5

    # Load test information
    match_list = read_match_list_from_adj_matrix(
        "tests/test_case_files/adj_matrix_1.txt"
    )
    f_in = open("tests/test_case_files/depthOnlyEP_1.txt", "r")
    line = f_in.readline()
    true_prob = float(line)

    # Call function
    prob, prob_error = probability(
        match_list,
        "player_0",
        "player_1",
        force_mode="average",
        model_name="depth_only",
    )

    # Compare results
    if np.abs(true_prob - prob) > deviations_threshold * prob_error:
        assert False

    # All good
    assert True


def test_luckOnly_probability():
    # How many times away from the claimed error we will allow
    deviations_threshold = 5

    # Load test information
    match_list = read_match_list_from_adj_matrix(
        "tests/test_case_files/adj_matrix_1.txt"
    )
    f_in = open("tests/test_case_files/luckOnlyEP_1.txt", "r")
    line = f_in.readline()
    true_prob = float(line)

    # Call function
    prob, prob_error = probability(
        match_list, "player_0", "player_1", force_mode="average", model_name="luck_only"
    )

    # Compare results
    if np.abs(true_prob - prob) > deviations_threshold * prob_error:
        assert False

    # All good
    assert True


def test_depthAndLuckOnly_probability():
    # How many times away from the claimed error we will allow
    deviations_threshold = 5

    # Load test information
    match_list = read_match_list_from_adj_matrix(
        "tests/test_case_files/adj_matrix_1.txt"
    )
    f_in = open("tests/test_case_files/depthAndLuckEP_1.txt", "r")
    line = f_in.readline()
    true_prob = float(line)

    # Call function
    prob, prob_error = probability(
        match_list,
        "player_0",
        "player_1",
        force_mode="average",
        model_name="depth_and_luck",
    )

    # Compare results
    if np.abs(true_prob - prob) > deviations_threshold * prob_error:
        assert False

    # All good
    assert True



# Parse the various supported file formats into a sparse adjecency matrix and list of labels
# TODO: Port to c++ to read large files faster
from scipy.sparse import lil_matrix
import networkx as nx
import numpy as np
from os.path import isfile


def read_match_list_from_match_list(filename):
    """Read list of matches from a file where each line represents a match in the format "winner loser".

    :param filename: Input filename
    :type filename: str
    :raises AssertionError: If a line is not in "winner loser" format. Labels should not themselves include spaces.
    """
    # Store the tuples of winner and loser indices since we first need to find the number of players
    match_list = []

    f = open(filename, "r")
    for line in f.readlines():
        line_split = line.split()
        if not len(line_split) == 2:
            raise AssertionError(
                f'"{line.strip()}" does not have 2 space-separated entries. \
Format as "winner loser" (make sure the names don\'t have spaces)'
            )

        match_list.append({"winner": line_split[0], "loser": line_split[1]})

    return match_list


def read_match_list_from_gml(filename):
    """Read list of matches as the edges in a gml network file.

    :param filename: Input filename
    :type filename: str
    :return: List of matches, each represented by a dict of the winner and loser.
    :rtype: list
    """
    # Read in with networkX
    G = nx.read_gml(filename)
    edges = list(G.edges())

    # Convert to our list of dicts format
    match_list = []
    for edge in edges:
        match_list.append({"winner": edge[0], "loser": edge[1]})

    return match_list


def read_match_list_from_adj_matrix(filename):
    """Read list of matches from an adjacency matrix which counts the number of times each player beats another. In lieu of labels, assign in the form player_i.

    :param filename: Input filename
    :type filename: str
    :return: List of matches, each represented by a dict of the winner and loser.
    :rtype: list
    """
    adj_matrix = np.loadtxt(filename, dtype=int)
    n = len(adj_matrix)

    match_list = []
    for i in range(n):
        for j in range(n):
            # Give each player a label from player_0 to player_(n-1) depending on where they appear in the given matrix
            player_i_label = f"player_{i}"
            player_j_label = f"player_{j}"
            for t in range(adj_matrix[i, j]):  # Add the number of matches which occured
                match_list.append({"winner": player_i_label, "loser": player_j_label})

    return match_list


def read_match_list(filename):
    """Read list of matches from a file, attempting to detect the appropriate file format among gml, match_list, or adjacency_matrix formats.

    :param filename: Input filename
    :type filename: str
    :return: List of matches, each represented by a dict of the winner and loser.
    :rtype: list
    """
    # First check that the file exists
    f_in = open(filename)

    # Attempt to read the file in the other of: gml, match_list, adj_matrix
    try:
        return read_match_list_from_gml(filename)
    except Exception:
        pass
    try:
        return read_match_list_from_match_list(filename)
    except Exception:
        pass
    try:
        return read_match_list_from_adj_matrix(filename)
    except Exception:
        pass

    # If all methods have failed, raise error
    raise AssertionError(
        f"Unable to parse {filename} as a gml, match list, or adjacency matrix."
    )



# Typical use case functions

import numpy as np
import pandas as pd
import warnings
import ranking
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde
from . import util


def scores(
    match_list,
    model_name="depth_and_luck",
    num_samples=5000,
    num_chains=4,
    force_mode=None,
):
    """Get fitted scores of players in a given model.

    :param match_list: List of matches, each represented by a dict of the winner and loser.
    :type match_list: list
    :param model_name: Model used for fitting. Defaults to 'depth_and_luck'. Options: {‘depth_and_luck’, ‘depth_only’, ‘luck_only’, ‘logistic_prior’}.
    :type model_name: str, optional
    :param num_samples: Number of samples used per chain for MCMC sampling, defaults to 5000
    :type num_samples: int, optional
    :param num_chains: Number of chains used for MCMC sampling, defaults to 4
    :type num_chains: int, optional
    :param force_mode: Optionally force the point estimate mode to be either 'average' or 'MAP', otherwise defaults to 'MAP' when more than 1000 players are present.
    :type force_mode: str or None, optional
    :return: pandas DataFrame with columns of the labels, inferred scores, and score errors for each player.
    :rtype: DataFrame
    """

    # Compute the number of participants present in match_list
    string_indices_dict = util.get_string_indices_dict(match_list)
    n = len(string_indices_dict)

    # Check the model_name is supported
    if model_name not in [
        "depth_and_luck",
        "depth_only",
        "luck_only",
        "logistic_prior",
    ]:
        raise AssertionError(f"model_name={model_name} not supported.")

    # Whether to find the expected values of the scores by sampling or use the MAP iteration method
    mode = "average"  # 'average' or 'MAP'
    if force_mode == None:
        if n < 1000:  # Prefer the sampling method if the data set is small enough
            mode = "average"
        else:  # If the data set is too large, default to using the simple iterative method
            warnings.warn(
                f"Data set contains {n} unique players, sampling may be slow for n > 1000.\n\
                        Defaulting to fast iterative method to find MAP estimates of the logistic_prior model.\n\
                            Can override this behavior with force_sampling=True."
            )
            mode = "MAP"
    elif force_mode == "average" or "MAP":
        mode = force_mode
    else:
        raise AssertionError(
            f"force_mode={force_mode} is not a valid option. \
                             Options are [None, 'average','MAP']"
        )

    if mode == "average":
        # Get samples from the posterior of this model fitted to this data set
        samples_df = ranking.samples(
            match_list,
            model_name=model_name,
            num_chains=num_chains,
            num_samples=num_samples,
        )

        # dict to be converted into scores DataFrame (more efficient to build this way)
        scores_df_dict = {"label": [], "score": [], "score_error": []}

        for column in samples_df.columns:  # Iterate over all sampled values
            if (
                column[-6:] == "_score"
            ):  # Check that this is a sampled score (not a parameter)
                # Extract the label name
                label = column[:-6]

                # Approximate the expected a posterori value and error by the sample mean and error
                sampled_values = samples_df[column]
                score_estimate = np.mean(sampled_values)
                score_error = np.std(sampled_values) / np.sqrt(len(sampled_values))

                # Write row to end of DataFrame
                scores_df_dict["label"].append(label)
                scores_df_dict["score"].append(score_estimate)
                scores_df_dict["score_error"].append(score_error)

        # Convert to DataFrame
        scores_df = pd.DataFrame(scores_df_dict)

        # Sort results by the estimated scores in descending order
        scores_df.sort_values(by="score", inplace=True, ascending=False)
        scores_df.reset_index(drop=True, inplace=True)

        # Set the index to start at 1 to nicely show the rank
        scores_df.index = range(1, len(scores_df) + 1)

        return scores_df
    else:
        # TODO: make logistic_prior_MAP faster
        MAP_scores, MAP_score_errors = ranking.logistic_prior_MAP(match_list)

        # Package information into a DataFrame
        labels = list(string_indices_dict.keys())
        scores_df_dict = {
            "label": labels,
            "score": MAP_scores,
            "score_error": MAP_score_errors,
        }

        # Convert to DataFrame
        scores_df = pd.DataFrame(scores_df_dict)

        # Sort results by the estimated scores in descending order
        scores_df.sort_values(by="score", inplace=True, ascending=False)
        scores_df.reset_index(drop=True, inplace=True)

        # Set the index to start at 1 to nicely show the rank
        scores_df.index = range(1, len(scores_df) + 1)

        return scores_df


def ranks(
    match_list,
    model_name="depth_and_luck",
    num_samples=5000,
    num_chains=4,
    force_mode=None,
):
    """Find the rankings of the players according to a specified model.

    :param match_list: List of matches, each represented by a dict of the winner and loser.
    :type match_list: list
    :param model_name: Model used for fitting. Defaults to 'depth_and_luck'. Options: {‘depth_and_luck’, ‘depth_only’, ‘luck_only’, ‘logistic_prior’}.
    :type model_name: str, optional
    :param num_samples: Number of samples used per chain for MCMC sampling, defaults to 5000
    :type num_samples: int, optional
    :param num_chains: Number of chains used for MCMC sampling, defaults to 4
    :type num_chains: int, optional
    :param force_mode: Optionally force the point estimate mode to be either 'average' or 'MAP', otherwise defaults to 'MAP' when more than 1000 players are present.
    :type force_mode: str or None, optional
    :return: Ranked list of the players in descending order of strength.
    :rtype: list
    """
    scores_df = scores(
        match_list,
        model_name=model_name,
        num_samples=num_samples,
        num_chains=num_chains,
        force_mode=force_mode,
    )

    return list(scores_df["label"])


def probability(
    match_list,
    winner_label,
    loser_label,
    model_name="depth_and_luck",
    num_samples=5000,
    num_chains=4,
    force_mode=None,
):
    """Inferred probability that one player will beat another, according to match_list data.

    :param match_list: List of matches, each represented by a dict of the winner and loser.
    :type match_list: list
    :param winner_label: Label of the desired winner
    :type winner_label: str
    :param loser_label: Label of the desired loser
    :type loser_label: str
    :param model_name: Model used for fitting. Defaults to 'depth_and_luck'. Options: {‘depth_and_luck’, ‘depth_only’, ‘luck_only’, ‘logistic_prior’}.
    :type model_name: str, optional
    :param num_samples: Number of samples used per chain for MCMC sampling, defaults to 5000
    :type num_samples: int, optional
    :param num_chains: Number of chains used for MCMC sampling, defaults to 4
    :type num_chains: int, optional
    :param force_mode: Optionally force the point estimate mode to be either 'average' or 'MAP', otherwise defaults to 'MAP' when more than 1000 players are present.
    :type force_mode: str or None, optional
    :raises AssertionError: If winner_label or loser_label is not present in match_list.
    :return: Tuple of the inferred probability and the error in the estimation.
    :rtype: tuple
    """
    # Compute the number of participants present in match_list
    string_indices_dict = util.get_string_indices_dict(match_list)
    n = len(string_indices_dict)

    # Check the model_name is supported
    if model_name not in [
        "depth_and_luck",
        "depth_only",
        "luck_only",
        "logistic_prior",
    ]:
        raise AssertionError(f"model_name={model_name} not supported.")

    # Check that the desired winner and loser are present in the data set
    if winner_label not in string_indices_dict.keys():
        raise AssertionError(f"{winner_label} not found in match_list.")
    if loser_label not in string_indices_dict.keys():
        raise AssertionError(f"{loser_label} not found in match_list.")

    # Whether to find the expected values of the scores by sampling or use the MAP iteration method
    mode = "average"  # 'average' or 'MAP'
    if force_mode == None:
        if n < 1000:  # Prefer the sampling method if the data set is small enough
            mode = "average"
        else:  # If the data set is too large, default to using the simple iterative method
            warnings.warn(
                f"Data set contains {n} unique players, sampling may be slow for n > 1000.\n\
                        Defaulting to fast iterative method to find MAP estimates of the logistic_prior model.\n\
                            Can override this behavior with force_sampling=True."
            )
            mode = "MAP"
    elif force_mode == "average" or "MAP":
        mode = force_mode
    else:
        raise AssertionError(
            f"force_mode={force_mode} is not a valid option. \
                             Options are [None, 'average','MAP']"
        )

    # Use sampling if the data set is not too large or the user has specified to
    if mode == "average":
        # Get samples from the posterior of this model fitted to this data set
        samples_df = ranking.samples(
            match_list,
            model_name=model_name,
            num_chains=num_chains,
            num_samples=num_samples,
        )

        # Get the total number of samples taken over all chains
        total_num_samples = len(samples_df)

        # Get the score differences between the relevant players
        # Need to add '_score' to the label for its column name
        sampled_score_differences = np.array(
            samples_df[f"{winner_label}_score"] - samples_df[f"{loser_label}_score"]
        )

        # Get parameter values. Depending on the model can set to different default values.
        match model_name:
            case "depth_and_luck":
                alphas = np.array(samples_df["alpha"])
                betas = np.array(samples_df["beta"])
            case "depth_only":
                alphas = np.zeros(total_num_samples)
                betas = np.array(samples_df["beta"])
            case "luck_only":
                alphas = np.array(samples_df["alpha"])
                betas = 100 * np.ones(
                    total_num_samples
                )  # We use this to approximate a true step function
            case "logistic_prior":
                alphas = np.zeros(total_num_samples)
                betas = np.ones(total_num_samples)
            case _:
                alphas = np.zeros(total_num_samples)
                betas = np.ones(total_num_samples)

        # Compute the probability of winning according to each sample
        probabilities = util.score_function(
            sampled_score_differences, alpha=alphas, beta=betas
        )

        # Probability is the mean over the samples
        probability = np.mean(probabilities)
        # Can estimate the error in the point estimate by the standard error
        probability_error = np.std(probabilities) / np.sqrt(total_num_samples)

        return (probability, probability_error)
    else:  # If the data set is too large, default to using the simple iterative method
        # TODO: make logistic_prior_MAP faster
        MAP_scores, MAP_score_errors = ranking.logistic_prior_MAP(match_list)

        winner_index = string_indices_dict[winner_label]
        loser_index = string_indices_dict[loser_label]
        score_difference = MAP_scores[winner_index] - MAP_scores[loser_index]
        # Adding the errors in quadrature as if they were independent
        score_difference_error = np.sqrt(
            MAP_score_errors[winner_index] ** 2 + MAP_score_errors[loser_index] ** 2
        )

        # Point estimate of the probability
        probability = util.score_function(score_difference, alpha=0, beta=1)
        # Estimate the error by changing he score difference by the score_error
        probability_upper = util.score_function(
            score_difference + score_difference_error, alpha=0, beta=1
        )
        probability_lower = util.score_function(
            score_difference - score_difference_error, alpha=0, beta=1
        )
        probability_error = (probability_upper - probability_lower) / 2

        return (probability, probability_error)


def depth_and_luck(
    match_list, model_name="depth_and_luck", num_samples=5000, num_chains=4
):
    """Get expected values of the depth and luck model parameters for a match_list by sampling. Returns appropriate stand-in parameters if model_name does not vary those parameters.

    :param match_list: List of matches, each represented by a dict of the winner and loser.
    :type match_list: list
    :param model_name: Model used for fitting. Defaults to 'depth_and_luck'. Options: {‘depth_and_luck’, ‘depth_only’, ‘luck_only’, ‘logistic_prior’}.
    :type model_name: str, optional
    :param num_samples: Number of samples used per chain for MCMC sampling, defaults to 5000
    :type num_samples: int, optional
    :param num_chains: Number of chains used for MCMC sampling, defaults to 4
    :type num_chains: int, optional
    :return: dict of depth (beta) and luck (alpha)
    :rtype: dict
    """
    # Get samples from the posterior of this model fitted to this data set
    samples_df = ranking.samples(
        match_list,
        model_name=model_name,
        num_chains=num_chains,
        num_samples=num_samples,
    )

    # Get the total number of samples taken over all chains
    total_num_samples = len(samples_df)

    # Estimate parameters by the sample means.
    # If a parameter is not sampled over return its appropriate fixed values
    # Note that for small data sets the expected value of beta can be very large,
    # since the first moment of the prior is unbounded: \int d\beta \beta P(\beta) = \infty
    match model_name:
        case "depth_and_luck":
            alphas = np.array(samples_df["alpha"])
            betas = np.array(samples_df["beta"])
        case "depth_only":
            alphas = np.zeros(total_num_samples)
            betas = np.array(samples_df["beta"])
        case "luck_only":
            alphas = np.array(samples_df["alpha"])
            betas = 100 * np.ones(
                total_num_samples
            )  # We use this to approximate a true step function
        case "logistic_prior":
            alphas = np.zeros(total_num_samples)
            betas = np.ones(total_num_samples)
        case _:
            alphas = np.zeros(total_num_samples)
            betas = np.ones(total_num_samples)

    alpha = np.mean(alphas)
    beta = np.mean(betas)
    return {"luck": alpha, "depth": beta}


def draw_depth_and_luck_posterior(match_list, num_samples=5000, num_chains=4):
    """Draw the posterior distribution of the luck and depth parameters from sampled values of the depth_and_luck model.

    :param match_list: List of matches, each represented by a dict of the winner and loser.
    :type match_list: list
    :param num_samples: Number of samples used per chain for MCMC sampling, defaults to 5000
    :type num_samples: int, optional
    :param num_chains: Number of chains used for MCMC sampling, defaults to 4
    :type num_chains: int, optional
    """
    # We can only use the depth_and_luck full model for this purpose
    samples_df = ranking.samples(
        match_list,
        model_name="depth_and_luck",
        num_chains=num_chains,
        num_samples=num_samples,
    )

    # Get the parameter samples
    alphas = samples_df["alpha"]
    betas = samples_df["beta"]

    # Show results uniform in theta = arctan(beta / 4) in order to show beta from 0 to infty
    alpha_spacing = 0.0025  # Setting the resolution
    theta_spacing = 0.005  # Setting the resolution
    # Define all possible values of alpha, beta, and theta (transformed beta) where we may evaluate the density estimate
    all_alphas = np.arange(0, 1, alpha_spacing)
    all_thetas = np.arange(0, np.pi / 2, theta_spacing)
    all_betas = util.theta_to_beta(all_thetas)

    # Restrict to the alphas and betas within the range of observed alphas and betas, this is where the grid will be computed
    compute_alphas = []
    for alpha in all_alphas:
        if alphas.min() < alpha and alpha < alphas.max():
            compute_alphas.append(alpha)
    compute_alphas = np.array(compute_alphas)
    compute_betas = []
    for beta in all_betas:
        if betas.min() < beta and beta < betas.max():
            compute_betas.append(beta)
    compute_betas = np.array(compute_betas)

    # Create (alpha,theta) pairs where we want to evaluate
    X, Y = np.meshgrid(compute_alphas, util.beta_to_theta(compute_betas))
    alpha_theta_pairs = np.vstack([X.ravel(), Y.ravel()])

    # Plot the samples and the Gaussian kernel smoothing (performed in (alpha,theta) space)
    thetas = util.beta_to_theta(betas)  # Compute the corresponding theta values

    # So that the heat map does not get cut off at alpha = 0, we mirror the points across
    # the alpha = 0 axis for density estimate calculation
    alphas_mirrored = []
    thetas_mirrored = []
    for i in range(len(alphas)):
        alphas_mirrored.append(alphas[i])
        thetas_mirrored.append(thetas[i])
        if alphas[i] < 0.03:  # Add an extra mirrored version
            alphas_mirrored.append(-alphas[i])
            thetas_mirrored.append(thetas[i])
    values = np.vstack([alphas_mirrored, thetas_mirrored])

    # Compute the gaussian kernel estimate
    kernel = gaussian_kde(values)
    kernal_vals = kernel(alpha_theta_pairs).T
    Z = np.reshape(kernal_vals, X.shape)

    # Plot the kernel estimate as a heatmap
    fig, ax = plt.subplots()
    ax.imshow(
        np.rot90(Z),
        cmap=plt.cm.gist_earth_r,
        extent=[np.min(thetas), np.max(thetas), np.min(alphas), np.max(alphas)],
    )

    # Plot the sampled points. Use less opacity when over 1000 sampled points.
    dot_alpha = min(1, float(1000) / len(alphas))
    ax.plot(thetas, alphas, "k.", markersize=2, alpha=dot_alpha)

    # Write in the tick values with the appropriate values of beta (ticks are uniform in theta)
    xtick_vals = []
    for x in np.arange(0, np.pi / 2, 0.1):
        if np.min(thetas) < x and x < np.max(thetas):
            xtick_vals.append(x)
    xtick_vals = np.array(xtick_vals)
    ax.set_xticks(xtick_vals)
    ax.set_xticklabels([f"{beta:.2f}" for beta in util.theta_to_beta(xtick_vals)])

    # Other plotting information
    ax.set_xlabel(r"$\beta$ (depth)")
    ax.set_ylabel(r"$\alpha$ (luck)")
    ax.set_xlim([np.min(thetas), np.max(thetas)])
    ax.set_ylim([np.min(alphas), np.max(alphas)])
    ax.set_aspect(1)



# __init__.py

__version__ = "1.0.0"

# These are imported in order listed
from ranking.parsing import *
from ranking.sampling import *
from ranking.fast_iteration import *
from ranking.results import *



import numpy as np


def beta_to_theta(beta):
    """Convert beta to theta

    :param beta: depth parameter
    :type beta: float
    :return: theta (angle of score function at 0)
    :rtype: float
    """
    return np.arctan(beta / 4)


def theta_to_beta(theta):
    """Convert theta to beta

    :param theta: (angle of score function at 0)
    :type theta: float
    :return: beta depth parameter
    :rtype: float
    """
    return 4 * np.tan(theta)


def sigmoid(x):
    """Sigmoid function, overflow protected and vectorized.

    :param x: input
    :type x: float
    :return: sigmoid
    :rtype: float
    """
    return np.where(x > 0, 1 / (1 + np.exp(-x)), np.exp(x) / (np.exp(x) + 1))


"""Score function f(s) for luck and depth parameter values.

    Args:
        s (float): score difference
        alpha (float, optional): luck parameter. Defaults to 0.
        beta (float, optional): depth parameter. Defaults to 1.

    Returns:
        float: score function value
    """


def score_function(s, alpha=0.0, beta=1.0):
    """Score function f(s) for luck and depth parameter values.

    :param s: score difference
    :type s: float
    :param alpha: luck parameter, defaults to 0.0
    :type alpha: float, optional
    :param beta: depth parameter, defaults to 1.0
    :type beta: float, optional
    :return: score function value
    :rtype: float
    """
    return 0.5 * alpha + (1 - alpha) * sigmoid(beta * s)


def get_string_indices_dict(match_list):
    """Generate a dict which maps each label found to an index in the order they appear.

    :param match_list: List of matches, each represented by a dict of the winner and loser.
    :type match_list: list
    :return: dict that associates to each found label a unique index.
    :rtype: dict
    """
    string_indices_dict = {}  # Dictionary of {string: index} for quick access

    for match in match_list:
        # Find (or assign) indices to the new labels
        winner_label = match["winner"]
        loser_label = match["loser"]
        if winner_label not in string_indices_dict.keys():
            string_indices_dict.update({winner_label: len(string_indices_dict)})
        if loser_label not in string_indices_dict.keys():
            string_indices_dict.update({loser_label: len(string_indices_dict)})

    return string_indices_dict


def make_match_list_hashable(match_list):
    """Turn a match_list into a hashable flattened tuple for caching.

    :param match_list: List of matches, each represented by a dict of the winner and loser.
    :type match_list: list
    :return: Flattened version of match_list
    :rtype: tuple
    """
    match_list_flat = []
    for match in match_list:
        match_list_flat.append(match["winner"])
        match_list_flat.append(match["loser"])
    return tuple(match_list_flat)


def undo_make_match_list_hashable(match_list_hashable):
    """Undo the make_match_list_hashable operation to recover match_list

    :param match_list_hashable: Flattened version of match_list
    :type match_list_hashable: tuple
    :raises AssertionError: If match_list_hashable does not have even length.
    :return: List of matches, each represented by a dict of the winner and loser.
    :rtype: list
    """
    if not len(match_list_hashable) % 2 == 0:
        raise AssertionError(
            f"match_list_hashable does not have an even number of entries."
        )

    match_list = []
    match_index = 0
    while match_index < len(match_list_hashable):
        # Convert the next two entries into a match
        match = {
            "winner": match_list_hashable[match_index],
            "loser": match_list_hashable[match_index + 1],
        }
        match_list.append(match)
        # Increment to process next pair
        match_index += 2

    return match_list



# Use pystan to take samples from the various models
import stan
import numpy as np
from functools import lru_cache
import warnings
import pandas as pd
from . import util  # Local import

# STAN model strings
depth_and_luck_STAN_string = """
data {
  int<lower = 0> N;                             // Number of players
  int<lower = 0> M;                             // Number of unique pairings
  array[M] int<lower=1, upper = N> player1;     // The index of "player1" in each pairing
  array[M] int<lower=1, upper = N> player2;     // The index of "player2" in each pairing
  array[M] real<lower = 0> A12;                 // Number of wins recorded by player1 against player2
}
parameters {
  vector[N] scores;                             // Score of each player
  real<lower = 0> beta;                         // Depth of competition
  real<lower = 0, upper = 1> alpha;             // Luck 
}
model {
  target += log(1/(4*(1 + beta^2/16)));         // P(beta), cauchy prior of width 4 (uniform over theta)
  for( k in 1:N){                               // P(s_i), prior on scores
    target += -(scores[k])^2;                   // Prior has variance 1/2 so that the difference between two scores is a Gaussian of width 1
  }
  for(i in 1:M){                                // P(A|s_i,alpha,beta), model likelihood
    real ds = scores[player1[i]] - scores[player2[i]];  // Score difference 
    target += A12[i]*log(alpha*0.5+(1-alpha)*1/(1+e()^(-beta*ds))); // Probability of times that player1 beat player2
  }
}
"""

depth_only_STAN_string = """
data {
  int<lower = 0> N;                             // Number of players
  int<lower = 0> M;                             // Number of unique pairings
  array[M] int<lower=1, upper = N> player1;     // The index of "player1" in each pairing
  array[M] int<lower=1, upper = N> player2;     // The index of "player2" in each pairing
  array[M] real<lower = 0> A12;                 // Number of wins recorded by player1 against player2
}
parameters {
  vector[N] scores;                             // Score of each player
  real<lower = 0> beta;                         // Depth of competition
}
model {
  target += log(1/(4*(1 + beta^2/16)));         // P(beta), cauchy prior of width 4 (uniform over theta)
  for( k in 1:N){                               // P(s_i), prior on scores
    target += -(scores[k])^2;                   // Prior has variance 1/2 so that the difference between two scores is a Gaussian of width 1
  }
  for(i in 1:M){                                // P(A|s_i,beta), model likelihood
    real ds = scores[player1[i]] - scores[player2[i]];  // Score difference 
    target += A12[i]*log(1/(1+e()^(-beta*ds))); // Probability of times that player1 beat player2
  }
}
"""

luck_only_STAN_string = """
data {
  int<lower = 0> N;                             // Number of players
  int<lower = 0> M;                             // Number of unique pairings
  array[M] int<lower=1, upper = N> player1;     // The index of "player1" in each pairing
  array[M] int<lower=1, upper = N> player2;     // The index of "player2" in each pairing
  array[M] real<lower = 0> A12;                 // Number of wins recorded by player1 against player2
}
parameters {
  vector[N] scores;                             // Score of each player
  real<lower = 0, upper = 1> alpha;             // Luck 
}
model {
  for( k in 1:N){                               // P(s_i), prior on scores
    target += -(scores[k])^2;                   // Prior has variance 1/2 so that the difference between two scores is a Gaussian of width 1
  }
  for(i in 1:M){                                // P(A|s_i,alpha), model likelihood. We slightly fake this by using beta = 100 to retain a continuous (and very nearly equivalent) model
    real ds = scores[player1[i]] - scores[player2[i]];  // Score difference 
    target += A12[i]*log(alpha*0.5+(1-alpha)*1/(1+e()^(-100*ds))); // Probability of times that player1 beat player2
  }
}
"""

logistic_prior_STAN_string = """
data {
  int<lower = 0> N;                             // Number of players
  int<lower = 0> M;                             // Number of unique pairings
  array[M] int<lower=1, upper = N> player1;     // The index of "player1" in each pairing
  array[M] int<lower=1, upper = N> player2;     // The index of "player2" in each pairing
  array[M] real<lower = 0> A12;                 // Number of wins recorded by player1 against player2
}
parameters {
  vector[N] scores;                             // Score of each player
}
model {
  for( k in 1:N){                               // P(s_i), prior on scores
    target += log(e()^(-scores[k])/(1+e()^(-scores[k]))^2); // Logistic prior
  }
  for(i in 1:M){                                // P(A|s_i), model likelihood
    real ds = scores[player1[i]] - scores[player2[i]];  // Score difference 
    target += A12[i]*log(1/(1+e()^(-ds)));      // Probability of times that player1 beat player2
  }
}
"""

# STAN language strings associated to each named model
STAN_strings_dict = {
    "depth_and_luck": depth_and_luck_STAN_string,
    "depth_only": depth_only_STAN_string,
    "luck_only": luck_only_STAN_string,
    "logistic_prior": logistic_prior_STAN_string,
}


def get_STAN_data(match_list, string_indices_dict):
    """Convert the match_list into a format for our STAN models.

    :param match_list: List of matches, each represented by a dict of the winner and loser.
    :type match_list: list
    :param string_indices_dict: dict that associates a unique index to each found string.
    :type string_indices_dict: dict
    :return: representation of the match_list for STAN.
    :rtype: dict
    """
    n = len(string_indices_dict)  # Number of players

    # Initializing STAN_data variables to be built up
    m = 0  # Number of unique pairings
    player1 = []  # The index of "player1" in each pairing
    player2 = []  # The index of "player2" in each pairing
    A12 = []  # Number of wins recorded by player1 against player2

    # Map the pairings present in the data to an index in the STAN_data arrays
    pairing_indices_dict = {}

    for match in match_list:
        winner_index = string_indices_dict[match["winner"]]
        loser_index = string_indices_dict[match["loser"]]

        # Store a pairing as the tuple [hashable!] (winner_index,loser_index)
        pairing = (winner_index, loser_index)

        # Assign this pairing an index if haven't already and add entries to
        # the STAN_data variables to represent this pairing.
        if pairing not in pairing_indices_dict.keys():
            pairing_indices_dict[pairing] = len(pairing_indices_dict)
            # Participants
            player1.append(pairing[0] + 1)  # STAN indices are unary
            player2.append(pairing[1] + 1)  # STAN indices are unary
            # Win counts
            A12.append(0)
            # Number of edges
            m += 1

        # Increment the STAN_data variables
        A12[pairing_indices_dict[pairing]] += 1

    STAN_data = {
        "N": int(n),
        "M": int(m),  # Number of unique pairings
        "player1": player1,  # The index of "player1" in each pairing
        "player2": player2,  # The index of "player2" in each pairing
        "A12": A12,
    }  # Number of wins recorded by player1 against player2

    return STAN_data


@lru_cache(maxsize=16)  # Cache samples taken for reuse (uses memory)
def _samples(
    match_list_hashable,
    model_name="depth_and_luck",
    num_chains=4,
    num_samples=10000,
    **kwargs,
):
    # Convert the hashable match_list back into our usual form
    match_list = util.undo_make_match_list_hashable(match_list_hashable)

    # STAN model string to be passed to pystan
    STAN_model_string = STAN_strings_dict[model_name]

    # print(f"Using model {model_name}")

    # Assign an index to each string in the match list
    string_indices_dict = util.get_string_indices_dict(match_list)

    # Get data from each of the
    STAN_data = get_STAN_data(match_list, string_indices_dict)

    # Compile and sample from the model (with the given data) using pystan
    posterior = stan.build(STAN_model_string, data=STAN_data, **kwargs)
    fit = posterior.sample(num_chains=num_chains, num_samples=num_samples)

    # Convert the resulting samples into a DataFrame with the original labels
    # DataFrame containing the STAN output (useful for diagnostics)
    df_original = fit.to_frame()
    # Dictionary to be converted into the final DataFrame
    df_dict = {}

    # Include the sampled model parameters if they are present
    for param in ["alpha", "beta"]:
        if param in df_original.columns:
            df_dict[param] = df_original[param]

    for name, index in string_indices_dict.items():
        STAN_index = index + 1  # STAN is unary
        STAN_label = f"scores.{STAN_index}"
        if STAN_label in df_original.columns:
            df_dict[f"{name}_score"] = df_original[STAN_label]
        else:
            raise AssertionError(f"{STAN_label} not found in pystan output.")

    # Convert to DataFrame
    df = pd.DataFrame(df_dict)

    return df


# TODO: switch to a permanent cache between executables. Store samples in local directory?
# Should probably add the option to not cache samples as well
def samples(
    match_list, model_name="depth_and_luck", num_chains=4, num_samples=10000, **kwargs
):
    """Get MCMC samples from the model fit to a match_list.

    :param match_list: List of matches, each represented by a dict of the winner and loser.
    :type match_list: list
    :param model_name: Model used for fitting. Defaults to 'depth_and_luck'. Options: {‘depth_and_luck’, ‘depth_only’, ‘luck_only’, ‘logistic_prior’}.
    :type model_name: str, optional
    :param num_samples: Number of samples used per chain for MCMC sampling, defaults to 10000
    :type num_samples: int, optional
    :param num_chains: Number of chains used for MCMC sampling, defaults to 4
    :type num_chains: int, optional
    :return: pandas DataFrame containing sampled draws of scores and relevant parameters.
    :rtype: DataFrame
    """
    # Give warning if the data set is too large for fast HMC sampling
    n = len(util.get_string_indices_dict(match_list))
    if n > 1000:
        # TODO: update warning with whatever we end up calling the fast iteration method
        warnings.warn(
            f"Data set contains {n} unique players, sampling may be slow for n > 1000.\n\
                      Consider using fast iterative MAP estimation methods. \n\
                        (Only available for the logistic_prior model)."
        )

    # Make the match_list hashable so that we can cache samples with lru_cache
    match_list_hashable = util.make_match_list_hashable(match_list)

    return _samples(
        match_list_hashable,
        model_name=model_name,
        num_chains=num_chains,
        num_samples=num_samples,
        **kwargs,
    )



from ranking.util import *
from ranking.parsing import read_match_list_from_match_list
import numpy as np
import pytest

numerical_tolerance = 0.0001


def test_beta_to_theta():
    test_beta = 2
    beta_error = test_beta - theta_to_beta(beta_to_theta(test_beta))
    assert np.abs(beta_error) < numerical_tolerance


def test_theta_to_beta():
    test_theta = 1
    theta_error = test_theta - beta_to_theta(theta_to_beta(test_theta))
    assert np.abs(theta_error) < numerical_tolerance


def test_score_function_default():
    sig_error = score_function(2) - 0.880797  # From Mathematica
    assert np.abs(sig_error) < numerical_tolerance


def test_score_function():
    sig_error = score_function(0.2, 0.1, 10) - 0.842717  # From Mathematica
    assert np.abs(sig_error) < numerical_tolerance


def test_make_match_list_hashable():
    match_list = read_match_list_from_match_list("data/match_lists/dogs.txt")
    match_list_hashable = make_match_list_hashable(match_list)
    print(type(match_list_hashable))
    assert type(match_list_hashable) is tuple

    match_list_recovered = undo_make_match_list_hashable(match_list_hashable)
    assert match_list == match_list_recovered



from ranking.parsing import *
import pytest


def test_read_outcome_list_too_many_spaces():
    with pytest.raises(AssertionError):
        assert read_match_list_from_match_list(
            "tests/test_case_files/spaces_in_names.txt"
        )
